/*

Goober Portable Gameboy Emulator
Copyright (C) 2002 Ian Smith

This Program Is Free Software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include "misc.h"

#include "allegro.h"
#include "gbcpu.h"
#include "mem.h"
#include "io.h"
#include "lcdc.h"
#include "intr.h"
#include "gui.h"
#include "joypad.h"
#include "cart.h"

/* Memory banks -- temporary -- */
static unsigned char rom0[0x4000] = {0};
static unsigned char rom1[0x4000] = {0};
static unsigned char exram[0x2000] = {0};
static unsigned char wram0[0x1000] = {0};
static unsigned char wram1[0x1000] = {0};
static unsigned char hram[0x7F] = {0};

/* timer function to record lcdc_frames every second */
static unsigned int frames=0, fps=0, actual_frames=0;
static void fps_proc();

/* Throttling function */
static unsigned int throttle_counter=0;
static void throttle_proc();

/* general timer function */
static unsigned int seconds=0;
static unsigned int freeze_timer=0;
static void timer_proc();

/* Load a ROM file */
static int load_rom(const char *fname);

/* Execute a specific amount of cycles */
static int execute_cycles(int cycles);

/* Initialize and Finalize GooBer */
static int init();
static void cleanup();

/* display the GUI */
static int show_gui();

/* GUI Callbacks */
static int on_file_exit();
static int on_file_open();
static int on_file_snapshot();
static int on_gb_reset();
static int on_gb_registers();
static int on_gb_debug();

/* Display Registers */
static void display_regs();

/* Get a key, return 0 if none pressed */
static int get_key();

/* Display/clear a message */
static void message(const char *msg);
static void clear_msg();

/* Memory access handlers -- temporary -- */
static unsigned char rom1_read(unsigned short addr);
static void rom1_write(unsigned short addr, unsigned char value);

static unsigned char exram_read(unsigned short addr);
static void exram_write(unsigned short addr, unsigned char value);

static unsigned char wram1_read(unsigned short addr);
static void wram1_write(unsigned short addr, unsigned char value);

/* CPU information */
gbcpu_context cpu;

/* Cartridge information */
static cart_info cart;

/* text fonts */
static DATAFILE *gb_font;
static FONT *big_font;

/* flags */
static int exit_flag=0;
static int rom_loaded=0;
static int reg_display_on=0;
static int debug_mode_on=0;
static int debug_exec=0;

/* Bitmap to draw register display to */
static BITMAP *reg_display;

/* settings */
static int screen_w;
static int screen_h;
static int display_w;
static int display_h;
static int display_x;
static int display_y;
static int scale_display;
static int msg_delay=5;

static int msg_x, msg_y;

static const char *rom_path=".";
static const char *screen_path=".";

static unsigned long frames_rendered=0;
static unsigned long frames_displayed=0;

static unsigned long cpu_seconds=0;
static unsigned long cycle_counter=0;
static unsigned long lcdc_cycles=0;

int main(int argc, char *argv[]) 
{
  BITMAP *temp;
  char title[] = "GooBer 0.0.0a [" PLATFORM "]";
  int key=0, copy_line=0, ctr;
  int redraw=1, frame_count=0, vblank=0;
  RGB pal[256];
  BITMAP *buffer;
  BITMAP *bg;

  printf("\nGooBer Gameboy Emulator v0.0.0a - development\n");
  printf("Copyright 2002 Ian Smith, portions copyright gnuboy\n");
  printf("gnuboy website: http://gnuboy.unix-fu.org\n");
  printf("contact: robotbebop@yahoo.ca\n\n");
 
  if (!init()) {
    return 0;
  }


  rect(screen, display_x-1, display_y-1, display_x+display_w, display_y+display_h, 15);
  textprintf_centre(screen, font, display_x + (display_w/2), display_y-7, 15, "%s", title);
  set_window_title(title);

  memory_init();
  io_init();
  lcdc_init();
  joypad_init();

  cpu.read = mem_read;
  cpu.write = mem_write;
  intr_set_cpu(&cpu);

  gbcpu_reset(&cpu);

  set_wram0_source(wram0);
  set_wram1_source(wram1);
  set_wram1_handlers(wram1_read, wram1_write);
  set_hram_source(hram);

  /* set GUI callbacks */
  gui_set_file_callback(M_FILE_EXIT, on_file_exit);
  gui_set_file_callback(M_FILE_OPEN, on_file_open);
  gui_set_file_callback(M_FILE_SNAP, on_file_snapshot);
  gui_set_gb_callback(M_GAMEBOY_RESET, on_gb_reset);
  gui_set_gb_callback(M_GAMEBOY_REGS, on_gb_registers);

  /* display menu */
  gui_enable_menu(M_GAMEBOY, 0);
  gui_enable_menu(M_DEBUG, 0);

  gui_show_menu();

  if (rom_loaded) {
    gui_enable_menu(M_GAMEBOY, 1);
    gui_enable_menu(M_DEBUG, 1);
  }

  lcdc_buffer = screen;

  lcdc_offset(display_x, display_y);
  if (scale_display) lcdc_size(display_w, display_h);

  while (!exit_flag) {
    if (seconds >= msg_delay) {
      clear_msg();
    }

    key = get_key();
    switch(key&0xFF) {
    case 27: /* Display GUI */
      gui_show_menu();
      break;
      
    case 'p': /* Snapshot hotkey */
    case 'P':
      on_file_snapshot();
      break;

    case 'x': /* Exit hotkey */
    case 'X':
      on_file_exit();
      break;
      
    case 'o': /* Open hotkey */
    case 'O':
      on_file_open();
      break;
      
    case 'r': /* Reset hotkey */
    case 'R':
      if (rom_loaded)	on_gb_reset();
      break;
            
    case 'd': /* toggle debug mode */
    case 'D':
      debug_mode_on = (debug_mode_on ? 0 : 1);
      break;

    case 'n': /* Execute next instruction */
    case 'N':
      debug_exec = 1;
      break;

    case 'e': /* Registers hotkey */
    case 'E':
      if (rom_loaded) on_gb_registers();
      break;

    }


    if (rom_loaded) {
      joypad_update();


      if ((--frame_count) > 0) {
	redraw = FALSE;
      }
      else {
	redraw = TRUE;
	frame_count = throttle_counter;
	throttle_counter -= frame_count;
      }

      if (redraw) {

	if (!vblank) {
	  lcdc_change(MODE_OAMTRANSFER);
	  if (!execute_cycles(CYCLES_OAMTRANSFER)) break;

	  lcdc_change(MODE_DATATRANSFER);
	  if (!execute_cycles(CYCLES_DATATRANSFER)) break;

	  lcdc_change(MODE_HBLANK);
	  if (!execute_cycles(CYCLES_HBLANK)) break;
	  
	  lcdc_render_scanline();

	  if (lcdc_scanline()==144) {
	    lcdc_change(MODE_VBLANK);
	    vblank = 1;
	  }
	}
	else {	
	  if (!execute_cycles(CYCLES_VBLANK/10)) break;
	  
	  lcdc_render_scanline();

	  if (lcdc_scanline() == 0) {
	    vblank = 0;
	    lcdc_render();
	    frames++;

	  }

	}

      }

    }

    
  }


  lcdc_cleanup();
  memory_cleanup();
  joypad_cleanup();

  cleanup();

  printf("\nTotal virtual seconds in execution: %d\n", cpu_seconds);
  return 0;
}

END_OF_MAIN();

static void fps_proc()
{
  fps = frames;
  frames = 0;
  actual_frames = 0;
}
END_OF_STATIC_FUNCTION(fps_proc);

static void throttle_proc()
{
  throttle_counter++;
}
END_OF_STATIC_FUNCTION(throttle_proc);

static void timer_proc()
{
  if (freeze_timer) {
    seconds++;
    if (seconds == 60) {
      seconds = 0;
    }
  }
}
END_OF_STATIC_FUNCTION(timer_proc);
     
/* Load a ROM file */
static int load_rom(const char *fname)
{
  int res;

  res = cart_load(fname, &cart);

  if (!res) return 0;

  printf("%s: Loaded ROM: \"%s\"\n", __FILE__, cart.title);
  return 1;
}

static int execute_cycles(int cycles)
{
  int total=0;
  int ret=1;

  while (total < cycles) {
    
    if (!gbcpu_ex(&cpu)) {
      ret = 0;
      break;
    }

    total += cpu.cycles;
    timer_update(cpu.cycles);

    display_regs();
  }

  return ret;
}

static int init()
{
  int force_window;

  allegro_init();

  LOCK_VARIABLE(frames);
  LOCK_VARIABLE(fps);
  LOCK_VARIABLE(seconds);
  LOCK_VARIABLE(freeze_timer);
  LOCK_VARIABLE(actual_frames);
  LOCK_VARIABLE(throttle_counter);

  LOCK_FUNCTION(throttle_proc);
  LOCK_FUNCTION(fps_proc);
  LOCK_FUNCTION(timer_proc);

  /* read options from goober.cfg */
  set_config_file("goober.cfg");
 
  msg_delay = get_config_int("goober", "msg_delay", 3);

  /* get paths */
  rom_path = get_config_string("paths", "rom_path", "./");
  screen_path = get_config_string("paths", "screen_path", "./");

  /* get screen configuration */
  screen_w = get_config_int("screen", "screen_w", 320);
  screen_h = get_config_int("screen", "screen_h", 240);
  force_window = get_config_int("screen", "force_window", 0);

  /* get display configuration */
  scale_display = get_config_int("display", "scale_display", 1);
  display_w = get_config_int("display", "display_w", 160);
  display_h = get_config_int("display", "display_h", 144);

  display_x = (screen_w/2) - (display_w/2);
  display_y = (screen_h/2) - (display_h/2);

  msg_x = display_x;
  msg_y = display_y + display_h + 2;
  
  if (force_window) {
    if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, screen_w, screen_h, 0, 0)!=0) {
      printf("Panic: Cannot set video mode: %s\n", allegro_error);
      return 0;
    }
  }
  else {
    if (set_gfx_mode(GFX_AUTODETECT, screen_w, screen_h, 0, 0)!=0) {
      printf("Panic: Cannot set video mode: %s\n", allegro_error);
      return 0;
    }
  }

  /* install input drivers */
  install_keyboard();
  install_mouse();
  install_timer();

  install_int_ex(fps_proc, BPS_TO_TIMER(1));
  install_int_ex(timer_proc, BPS_TO_TIMER(1));
  install_int_ex(throttle_proc, BPS_TO_TIMER(60));

  /* set GUI background color */
  gui_bg_color = 1;

  /* Load the font */
  gb_font = load_datafile_object("goober.dat", "SMALL_FONT");
  big_font = font;
  font = (FONT*)gb_font->dat;

  /* create the register display bitmap */
  reg_display = create_bitmap(320, 44);
  clear(reg_display);

  return 1;
}


static void cleanup()
{

  destroy_bitmap(reg_display);
  remove_int(fps_proc);
  remove_int(timer_proc);
  remove_int(throttle_proc);

  remove_keyboard();
  remove_mouse();
  remove_timer();

  set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
  unload_datafile_object(gb_font);

  printf("Frames rendered: %d\nFrames displayed: %d\n", 
	 frames_rendered, frames_displayed);

}

static int on_file_exit()
{
  exit_flag=1;
  return 1;
}

static int on_file_open()
{
  char open_from[480]={0};
  int res;

  strcpy(open_from, rom_path);
  res = file_select_ex("Select ROM Image (*.gb)", open_from, "GB", 255, 300, 220);

  if (res) {
    if (load_rom(open_from)) {     
      on_gb_reset();
      gui_enable_menu(M_GAMEBOY, 1);
      gui_enable_menu(M_DEBUG, 1);
      rom_loaded=1;
      clear_msg();
      message("Loaded ROM");
    }
  }

  return 1;
}

static int on_file_snapshot()
{
  static int snap_count=0;
  char snap_name[480]={0};

  freeze_timer = 1;
  message("Saving snapshot to ... ");

  snap_count++;
  sprintf(snap_name, "%ssnapshot_%d.pcx", screen_path, snap_count);

/*   get_palette(pal); */
/*   save_pcx(snap_name, screen, pal);  */
  lcdc_capture(snap_name);

  clear_msg();
  message(snap_name);

  seconds=0;
  freeze_timer=0;

  return 1;
}

static int on_gb_reset()
{
  gbcpu_reset(&cpu);
  memory_reset();
  lcdc_reset();
  mbc_reset();
  display_regs();
  return 1;
}

static int on_gb_registers()
{
  if (reg_display_on)
    reg_display_on=0;
  else
    reg_display_on=1;

  display_regs();
  gui_gb_selected(M_GAMEBOY_REGS, reg_display_on);
  return 1;
}

static void display_regs()
{
  char ins[255]={0};

  unsigned short af = WORD(cpu.a, cpu.flags);
  unsigned short bc = WORD(cpu.b, cpu.c);
  unsigned short de = WORD(cpu.d, cpu.e);
  unsigned short hl = WORD(cpu.h, cpu.l);

  if (!reg_display_on) return;

  textprintf(reg_display, font, 0, 0, 15, "AF: %.4x  BC: %.4x  DE: %.4x  HL: %.4x", 
	     WORD(cpu.a, cpu.flags), WORD(cpu.b, cpu.c), WORD(cpu.d, cpu.e),
	     WORD(cpu.h, cpu.l)
	     );
  textprintf(reg_display, font, 0, 10, 15, "PC: %.4x  SP: %.4x", cpu.pc, cpu.sp);

  rectfill(reg_display, 0, 17, reg_display->w, 27, 0);
  gbcpu_da(&cpu, ins);
  textprintf(reg_display, font, 0, 17, 15, "%s", ins);

  blit(reg_display, screen, 0, 0, 0, 0, reg_display->w, reg_display->h);

}

static int get_key()
{
  if (keypressed()) 
    return readkey();
  else
    return 0;
}

/* Display/clear a message */
static void message(const char *msg)
{
  textout(screen, font, msg, msg_x, msg_y, 15);
}

static void clear_msg()
{
  rectfill(screen, msg_x, msg_y, 320, msg_y+10, 0);
}

/* Memory access handlers -- temporary -- */
static unsigned char rom1_read(unsigned short addr)
{
  return rom1[addr];
}

static void rom1_write(unsigned short addr, unsigned char value)
{
  /* do nothing ROM is _READ_ONLY_ durrgghhhh */
}

static unsigned char exram_read(unsigned short addr)
{
  return exram[addr];
}

static void exram_write(unsigned short addr, unsigned char value)
{
  exram[addr] = value;
}

static unsigned char wram1_read(unsigned short addr)
{
  return wram1[addr];
}

static void wram1_write(unsigned short addr, unsigned char value)
{
  wram1[addr] = value;
}


